"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const chance_1 = require("chance");
const utils = require("./utils");
const chatito = require('../parser/chatito');
const chance = new chance_1.Chance();
const chatitoFormatPostProcess = (data) => {
    const arr = data.reduce((accumulator, next, i, arrShadow) => {
        if (accumulator.length) {
            const lastWord = accumulator[accumulator.length - 1];
            if (lastWord.type === next.type && lastWord.type === 'Text') {
                accumulator[accumulator.length - 1] = {
                    type: lastWord.type,
                    value: (lastWord.value + next.value).replace(/\s+/g, ' ')
                };
            }
            else {
                accumulator.push(next);
            }
        }
        else if (next.value.trim()) {
            accumulator.push(next);
        }
        if (i === arrShadow.length - 1) {
            if (accumulator.length) {
                if (!accumulator[accumulator.length - 1].value.trim()) {
                    accumulator.pop();
                }
                accumulator[accumulator.length - 1] = Object.assign({}, accumulator[accumulator.length - 1], {
                    value: accumulator[accumulator.length - 1].value.replace(/\s+$/g, '')
                });
            }
        }
        return accumulator;
    }, []);
    if (arr.length) {
        arr[0] = Object.assign({}, arr[0], {
            value: arr[0].value.replace(/^\s+/, '')
        });
    }
    if (!arr.length) {
        throw new Error(`Some sentence generated an empty string. Can't map empty to an intent.`);
    }
    return arr;
};
const getVariationsFromEntity = (ed, entities, optional, cache) => __awaiter(this, void 0, void 0, function* () {
    const variationKey = ed.variation ? `#${ed.variation}` : '';
    const cacheKey = `${ed.type}-${ed.key}${variationKey}`;
    let cacheStats = cache.get(cacheKey);
    if (!cacheStats) {
        const counts = [];
        const maxCounts = [];
        const definedSentenceProbabilities = [];
        const indexesOfSentencesWithNullProbability = [];
        let sumOfTotalProbabilitiesDefined = 0;
        for (const c of ed.inner) {
            counts.push(new Map());
            let mxc = utils.maxSentencesForSentence(entities)(c);
            if (optional) {
                mxc++;
            }
            maxCounts.push(mxc);
            if (c.probability === null) {
                indexesOfSentencesWithNullProbability.push(definedSentenceProbabilities.length);
                definedSentenceProbabilities.push(null);
            }
            else {
                const prob = parseInt(c.probability || '', 10);
                if (!Number.isInteger(prob)) {
                    throw new Error(`Probability "${c.probability}" must be an integer value. At ${cacheKey}`);
                }
                if (prob < 1 || prob > 100) {
                    throw new Error(`Probability "${c.probability}" must be from 1 to 100. At ${cacheKey}`);
                }
                sumOfTotalProbabilitiesDefined += prob;
                definedSentenceProbabilities.push(prob);
            }
        }
        if (sumOfTotalProbabilitiesDefined && sumOfTotalProbabilitiesDefined > 100) {
            throw new Error(`The sum of sentence probabilities (${sumOfTotalProbabilitiesDefined}) for an entity can't be higher than 100%. At ${cacheKey}`);
        }
        const totalMaxCountsToShareBetweenNullProbSent = indexesOfSentencesWithNullProbability.map(i => maxCounts[i]).reduce((p, n) => (p || 0) + (n || 0)) || 0;
        const probabilities = definedSentenceProbabilities.map((p, i) => p === null
            ? (((maxCounts[i] * 100) / totalMaxCountsToShareBetweenNullProbSent) * (100 - sumOfTotalProbabilitiesDefined)) / 100
            : p);
        const currentEntityCache = { counts, maxCounts, optional, probabilities };
        cache.set(cacheKey, currentEntityCache);
        cacheStats = cache.get(cacheKey);
    }
    const optionalProb = 100 / (cacheStats.probabilities.length + 1);
    let sentenceIndex = chance.weighted(Array.from(cacheStats.probabilities.keys()), cacheStats.probabilities);
    if (cacheStats.optional && chance.bool({ likelihood: optionalProb })) {
        sentenceIndex = -1;
    }
    if (sentenceIndex === -1) {
        return [];
    }
    const sentence = ed.inner[sentenceIndex].sentence;
    let accumulator = [];
    const slotSynonyms = ed.type === 'SlotDefinition' && sentence.length === 1 && sentence[0].type === 'Alias';
    for (const t of sentence) {
        const slotsInSentenceKeys = new Set([]);
        if (t.type === 'Slot' || t.type === 'Alias') {
            const def = entities[t.type];
            const innerEntityKey = t.variation ? `${t.value}#${t.variation}` : t.value;
            const currentCache = slotsInSentenceKeys.has(innerEntityKey) ? cacheStats.counts[sentenceIndex] : new Map();
            slotsInSentenceKeys.add(innerEntityKey);
            const sentenceVariation = yield getVariationsFromEntity(def[innerEntityKey], entities, !!t.opt, currentCache);
            if (sentenceVariation.length) {
                const returnSentenceTokens = chatitoFormatPostProcess(sentenceVariation);
                for (const returnToken of returnSentenceTokens) {
                    if (slotSynonyms) {
                        returnToken.synonym = t.value;
                    }
                    if (t.type === 'Slot') {
                        if (def[innerEntityKey].args) {
                            returnToken.args = def[innerEntityKey].args;
                        }
                        returnToken.value = returnToken.value.trim();
                        returnToken.type = t.type;
                        returnToken.slot = t.value;
                    }
                    accumulator = accumulator.concat(returnToken);
                }
            }
        }
        else {
            accumulator = accumulator.concat(t);
        }
    }
    return accumulator;
});
exports.astFromString = (str) => chatito.parse(str);
exports.datasetFromString = (str, writterFn, importer, currentPath) => {
    const ast = exports.astFromString(str);
    return exports.datasetFromAST(ast, writterFn, importer, currentPath);
};
exports.getImports = (from, to, importer) => {
    const fileContent = importer(from, to);
    if (!fileContent || !fileContent.dsl) {
        throw new Error(`Failed importing ${to}`);
    }
    try {
        const importAst = exports.astFromString(fileContent.dsl);
        let outAst = [];
        importAst.forEach(ett => {
            if (ett.type === 'ImportFile' && ett.value) {
                outAst = [...outAst, ...exports.getImports(fileContent.filePath, ett.value, importer)];
            }
            else if (ett.type === 'AliasDefinition' || ett.type === 'SlotDefinition') {
                outAst = [...outAst, ett];
            }
        });
        return outAst;
    }
    catch (e) {
        throw new Error(`Failed importing ${to}. ${e.message} - ${JSON.stringify(e.location)}`);
    }
};
exports.datasetFromAST = (initialAst, writterFn, importHandler, currPath) => __awaiter(this, void 0, void 0, function* () {
    const operatorDefinitions = { Intent: {}, Slot: {}, Alias: {} };
    if (!initialAst || !initialAst.length) {
        return;
    }
    const importer = importHandler ? importHandler : () => ({ filePath: '', dsl: '' });
    const currentPath = currPath ? currPath : '';
    let ast = [...initialAst];
    initialAst.forEach(od => {
        if (od.type === 'ImportFile' && od.value) {
            ast = [...ast, ...exports.getImports(currentPath, od.value, importer)];
        }
    });
    ast.forEach(od => {
        let entity;
        if (od.type === 'IntentDefinition') {
            entity = operatorDefinitions.Intent;
        }
        else if (od.type === 'SlotDefinition') {
            entity = operatorDefinitions.Slot;
        }
        else if (od.type === 'AliasDefinition') {
            entity = operatorDefinitions.Alias;
        }
        else if (od.type === 'Comment' || od.type === 'ImportFile') {
            return;
        }
        else {
            throw new Error(`Unknown definition definition for ${od.type}`);
        }
        const odKey = od.variation ? `${od.key}#${od.variation}` : od.key;
        if (entity[odKey]) {
            throw new Error(`Duplicate definition for ${od.type} '${odKey}'`);
        }
        entity[odKey] = od;
    });
    const intentKeys = Object.keys(operatorDefinitions.Intent);
    if (!intentKeys || !intentKeys.length) {
        return;
    }
    for (const intentKey of intentKeys) {
        const maxPossibleCombinations = utils.maxSentencesForEntity(operatorDefinitions.Intent[intentKey], operatorDefinitions);
        let maxIntentExamples = maxPossibleCombinations;
        const entityArgs = operatorDefinitions.Intent[intentKey].args;
        let trainingN = maxIntentExamples;
        let testingN = 0;
        let generatedTrainingExamplesCount = 0;
        let generatedTestingExamplesCount = 0;
        if (entityArgs) {
            if (entityArgs.training) {
                trainingN = parseInt(entityArgs.training, 10);
                if (trainingN < 1) {
                    throw new Error(`The 'training' argument for ${intentKey} must be higher than 0.`);
                }
                if (entityArgs.testing) {
                    testingN = parseInt(entityArgs.testing, 10);
                    if (testingN < 1) {
                        throw new Error(`The 'testing' argument for ${intentKey} must be higher than 0.`);
                    }
                }
            }
            const intentMax = trainingN + testingN;
            if (intentMax > maxIntentExamples) {
                throw new Error(`Can't generate ${intentMax} examples. Max possible examples is ${maxIntentExamples}`);
            }
            else if (intentMax < maxIntentExamples) {
                maxIntentExamples = intentMax;
            }
        }
        const maxEx = maxIntentExamples;
        const globalCache = new Map();
        const collitionsCache = {};
        let duplicatesCounter = 0;
        while (maxIntentExamples) {
            const intentSentence = yield getVariationsFromEntity(operatorDefinitions.Intent[intentKey], operatorDefinitions, false, globalCache);
            const utterance = chatitoFormatPostProcess(intentSentence);
            const utteranceString = utterance.reduce((p, n) => p + n.value, '');
            if (!collitionsCache[utteranceString]) {
                collitionsCache[utteranceString] = true;
                const completedTraining = generatedTrainingExamplesCount >= trainingN;
                const completedTesting = generatedTestingExamplesCount >= testingN;
                let isTrainingExample = !completedTraining;
                if (!completedTraining && !completedTesting) {
                    isTrainingExample = Math.random() < 0.7;
                }
                writterFn(utterance, intentKey, isTrainingExample);
                maxIntentExamples--;
                if (isTrainingExample) {
                    generatedTrainingExamplesCount++;
                }
                else {
                    generatedTestingExamplesCount++;
                }
            }
            else {
                duplicatesCounter++;
                const maxDupes = maxPossibleCombinations * maxPossibleCombinations;
                const maxDupesLimit = Math.floor(maxPossibleCombinations / 2);
                if (duplicatesCounter > (maxPossibleCombinations > 10000 ? maxDupesLimit : maxDupes)) {
                    let m = `Too many duplicates while generating dataset! Looks like we have probably reached `;
                    m += `the maximun ammount of possible unique generated examples. `;
                    m += `The generator has stopped at ${maxEx - maxIntentExamples} examples for intent ${intentKey}.`;
                    console.warn(m);
                    maxIntentExamples = 0;
                }
            }
        }
    }
});
